from base64 import urlsafe_b64decode, urlsafe_b64encode
import json
import os
import re

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from . import _typing as typ
from .constants import INVALID_YUBICO_CERT_SHASUMS
from .enums import RequestType
from .exceptions import U2FInvalidDataException

BASE64URL = re.compile(br'^[-_a-zA-Z0-9]*=*$')


def get_random_challenge() -> bytes:
    return os.urandom(64)


def standardise_client_data(raw_client_data: str) -> str:
    try:
        if '{' not in raw_client_data:
            # Assume base64 encoded JSON.
            return websafe_decode(raw_client_data).decode('ascii')
        else:
            # Assume JSON string.
            return raw_client_data
    except ValueError:
        raise U2FInvalidDataException('Client data was an invalid string')


def load_client_data(raw_client_data: str) -> typ.Mapping[str, typ.Any]:
    if isinstance(raw_client_data, str):
        try:
            return json.loads(standardise_client_data(raw_client_data))
        except ValueError:
            raise U2FInvalidDataException('Client data was an invalid string')
    else:
        raise U2FInvalidDataException(
            'Client data is an unsupported type{!r}.'.format(
                type(raw_client_data)))


def validate_client_data(
        raw_client_data: str,
        request_type: RequestType,
        app_id: str,
        expected_challenge: str,
) -> str:
    standardised_client_data = standardise_client_data(raw_client_data)
    client_data = load_client_data(standardised_client_data)
    print(client_data)
    if client_data.get('typ', None) != request_type.value:
        raise U2FInvalidDataException('Invalid or missing request type')
    if client_data.get('origin', None) != app_id:
        raise U2FInvalidDataException('Invalid or missing origin')
    if client_data.get('challenge', None) != expected_challenge:
        raise U2FInvalidDataException('Invalid or missing challenge')
    # Valid client data falls through.
    return standardised_client_data


def sha_256(data: bytes) -> bytes:
    h = hashes.Hash(hashes.SHA256(), default_backend())
    h.update(data)
    return h.finalize()


def pop_bytes(data: bytearray, l: int) -> bytes:
    x = bytes(data[:l])
    del data[:l]
    return x


def fix_invalid_yubico_certs(der: bytes):
    # Some early certs have UNUSED BITS incorrectly set.
    # Fix only if they are one of the known bad
    if sha_256(der) in INVALID_YUBICO_CERT_SHASUMS:
        der = der[:-257] + b'\0' + der[-256:]
    return der


def parse_tlv_encoded_length(data: bytearray) -> int:
    # https://msdn.microsoft.com/en-us/library/windows/desktop/bb648641(v=vs.85).aspx
    # Starting at 1 because byte 0 is the 'tag'
    length = data[1] & 0x7f
    extended_length_flag = data[1] & 0x80
    if extended_length_flag:
        # The 7 low-bits of `length` indicate the number of bytes to read to
        #  determine the length
        true_length = 0
        # Offset at 2 because we started at 1; and we've already read the
        #  first byte.
        for byte in data[2:(2 + length)]:
            true_length = (true_length << 8) | byte
        # Return the length; plus the bytes we've already read.
        return 2 + length + true_length
    else:
        # High bit is unset; meaning the value in the lower 7 bits must be the
        #  length of the certicate.
        # Return that plus the 2 bytes we've already read.
        return 2 + length


def websafe_decode(data: typ.Union[bytes]) -> bytes:
    """Convert the URL Safe Base64 string into the bytes it represents."""
    if isinstance(data, str):
        data = data.encode('ascii')
    if not BASE64URL.match(data):
        raise ValueError('Invalid character(s)')
    data += b'=' * (-len(data) % 4)
    return urlsafe_b64decode(data)


def websafe_encode(data: typ.Union[bytes]) -> bytes:
    """Convert the given data into it's URL Safe Base64 representation."""
    if isinstance(data, str):
        data = data.encode('ascii')
    return urlsafe_b64encode(data).rstrip(b'=').decode('ascii')


class _AbstractAttribute():
    def __get__(self, obj, type):
        this_obj = obj if obj else type
        abc_class = None
        attr_name = None
        # Find ourselves in the MRO
        for cls in type.__mro__:
            for name, value in cls.__dict__.items():
                if value is self:
                    abc_class = cls
                    attr_name = name
        if abc_class is not None and attr_name is not None:
            raise NotImplementedError(
                '%r does not have the attribute %r (abstract from class %r)' %
                (this_obj, name, cls.__name__))
        else:
            raise NotImplementedError(
                '%r does not set the abstract attribute <unknown>' % this_obj)


def abstract_attribute() -> typ.Any:
    """
    An attribute that throws an error if it is not provided in a subclass.
    """
    # This function keeps MyPy happy
    return _AbstractAttribute()



import abc

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec

from . import _typing as typ
from .constants import U2F_TRANSPORT_EXTENSION_OID, U2F_V2
from .device import DeviceRegistration, device_as_client_dict
from .enums import RequestType, U2FTransport, U2FTransports
from .exceptions import U2FInvalidDataException, U2FStateException
from .utils import fix_invalid_yubico_certs, get_random_challenge, sha_256
from .utils import parse_tlv_encoded_length, pop_bytes, validate_client_data
from .utils import websafe_decode, websafe_encode


class U2FRegistrationManager(abc.ABC):

    REGISTRATION_SESSION_KEY = 'u2f_registration_challenge'

    def __init__(self, app_id: str) -> None:
        self.app_id = app_id

    @abc.abstractmethod
    def create_device_registration_model(
        self,
        *,
        version: str,
        app_id: str,
        key_handle: bytes,
        public_key: bytes,
        transports: U2FTransports
    ) -> DeviceRegistration:
        ...

    def create_registration_challenge(
        self,
        session: typ.MutableMapping[str, typ.Any],
        registered_devices: typ.Collection[DeviceRegistration] = (),
    ) -> typ.Mapping[str, typ.Any]:
        """
        Generate a challenge and return information for the registration step.

        This will generate a secure random challenge, placing it in the session
        object, and then return a JSON-safe object for use by the client to
        complete the challenge
        """
        challenge = websafe_encode(get_random_challenge())
        session[self.REGISTRATION_SESSION_KEY] = challenge
        return {
            'appId': self.app_id,
            'registerRequests': [{
                'version': U2F_V2,
                'challenge': challenge,
            }],
            'registeredKeys': [
                device_as_client_dict(key)
                for key in registered_devices
            ],
        }

    def process_registration_response(
        self,
        session: typ.MutableMapping[str, typ.Any],
        response_dict: typ.Mapping[str, str],
    ) -> DeviceRegistration:
        version = response_dict.get('version', '')
        challenge = session.pop(self.REGISTRATION_SESSION_KEY, None)
        if not challenge:
            raise U2FStateException('Session missing required key.')
        if version != U2F_V2:
            raise U2FInvalidDataException('Unsupported version given.')
        registration_data = self.verify_registration_data(
            response_dict, challenge)
        # We have now verified the registration request.
        return self.create_device_registration_model(
            version=version,
            app_id=self.app_id,
            key_handle=registration_data.key_handle,
            public_key=registration_data.public_key,
            transports=registration_data.get_supported_transports(),
        )

    def verify_registration_data(
        self,
        response_dict: typ.Mapping[str, str],
        challenge: str,
    ) -> 'RegistrationData':
        try:
            registration_data = RegistrationData.from_base64(
                response_dict.get('registrationData', ''))
        except (ValueError, IndexError) as e:
            raise U2FInvalidDataException('Invalid registration data.') from e
        # Client data comes in as base64(usually?), so we standardise it
        #  into a decoded *string*. We then take the hash of that string
        #  for the verification step.
        client_data = validate_client_data(
            response_dict.get('clientData', ''),
            RequestType.REGISTER,
            self.app_id,
            challenge,
        )
        challenge_param = sha_256(client_data.encode('utf-8'))
        app_param = sha_256(self.app_id.encode('idna'))
        registration_data.verify(app_param, challenge_param)
        return registration_data


class RegistrationData():

    @classmethod
    def from_base64(
        cls,
        base64_data: typ.Union[str, bytes],
    ) -> 'RegistrationData':
        return cls(websafe_decode(base64_data))  # type: ignore

    def __init__(self, data: bytes) -> None:
        # https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.pdf
        buf = bytearray(data)
        if buf.pop(0) != 0x05:
            raise U2FInvalidDataException(
                'Registration data has invalid magic byte')
        self.public_key = pop_bytes(buf, 65)
        self.key_handle = pop_bytes(buf, buf.pop(0))
        cert_len = parse_tlv_encoded_length(buf)
        self.certificate = fix_invalid_yubico_certs(pop_bytes(buf, cert_len))
        self.signature = bytes(buf)

    def get_x509_certificate(self) -> x509.Certificate:
        return x509.load_der_x509_certificate(
            self.certificate, default_backend())

    def verify(self, app_param: bytes, chal_param: bytes) -> None:
        # https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.pdf
        cert = self.get_x509_certificate()
        pubkey = cert.public_key()
        verifier = pubkey.verifier(self.signature, ec.ECDSA(hashes.SHA256()))

        verifier.update(
            b'\0'  # control byte
            + app_param
            + chal_param
            + self.key_handle
            + self.public_key,
        )
        try:
            verifier.verify()
        except InvalidSignature as e:
            raise U2FInvalidDataException(
                'Attestation signature is invalid') from e

    def get_supported_transports(
        self,
    ) -> U2FTransports:
        """Extract the transports this token supports from the certificate."""
        cert = self.get_x509_certificate()
        try:
            ext = cert.extensions.get_extension_for_oid(
                U2F_TRANSPORT_EXTENSION_OID)  # type: x509.Extension
        except x509.ExtensionNotFound:
            # Supported transports unknown. Spec indicates this must be `null`
            return None
        # Access the raw bytes in the extension field.
        transports_bitstring = ext.value.value
        # `transports_bitstring` is 4 bytes:
        assert transports_bitstring[0] == 0x03
        assert transports_bitstring[1] == 0x02
        unused_bits = transports_bitstring[2]
        transport_flags = transports_bitstring[3]
        # The last `unused_bits` should be unset. Make sure that they are
        transport_flags = (transport_flags >> unused_bits) << unused_bits
        return U2FTransport.from_byte(transport_flags)



from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Column, ForeignKey, Integer, LargeBinary, String
from sqlalchemy.orm import relationship

from ..enums import U2FTransport
from ..device import DeviceRegistration


db = SQLAlchemy()


class User(db.Model):
    __tablename__ = 'User'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    devices = relationship('Device', back_populates='user')

    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
       return "<User(%s, name='%s')>" % (self.id, self.name)


class Device(db.Model, DeviceRegistration):
    __tablename__ = 'Device'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('User.id'))
    user = relationship("User", back_populates="devices")

    """The U2F protocol version to use."""
    version = Column(String)
    """The appId used during registration."""
    app_id = Column(String)
    """The raw `keyHandle`."""
    key_handle = Column(LargeBinary)
    """The raw `pubKey`."""
    public_key = Column(LargeBinary)
    """An integer-encoded `U2FTransport`. Use `u2f_transports` instead."""
    transports = Column(Integer)
    """A 32-bit unsigned integer representing the last seen counter."""
    counter = Column(Integer)

    @property
    def u2f_transports(self):
        if self.transports < 0:
            return None
        return U2FTransport._from_internal_int(self.transports)

    @u2f_transports.setter
    def u2f_transports(
        self,
        transports,
    ):
        self.transports = U2FTransport._to_internal_int(transports)





class U2FException(Exception):
    pass


class U2FStateException(U2FException):
    """Raised when the system's state is incorrect for the desired action."""
    pass


class U2FInvalidDataException(U2FException):
    """
    Raised when data given by the caller is not valid; or fails verification.
    """
    pass



from . import _typing as typ
from .utils import abstract_attribute, websafe_encode


class DeviceRegistration():

    version = abstract_attribute()  # type: str
    app_id = abstract_attribute()  # type: str
    key_handle = abstract_attribute()  # type: bytes
    public_key = abstract_attribute()  # type: bytes
    counter = abstract_attribute()  # type: int

    # U2FTransports
    u2f_transports = abstract_attribute()


def device_as_client_dict(device: DeviceRegistration) -> typ.Dict[str, typ.Any]:
    transports = device.u2f_transports
    if transports is not None:
        out_transports = sorted([
            t.internal_name for t in transports
        ])  # type: typ.Optional[typ.Collection[str]]
    else:
        out_transports = None
    return {
        'version': device.version,
        'appId': device.app_id,
        # keyHandle must be base64 encoded
        'keyHandle': websafe_encode(device.key_handle),
        'transports': out_transports,
    }



import json
import pathlib

from flask import Flask, request, session, redirect

from .tables import db, User, Device
from .. import registration, verification

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////tmp/data.db'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
app.config['SECRET_KEY'] = 'WebAuthN would be nice one day.'
app.config['DEBUG'] = True

db.init_app(app)

@app.before_first_request
def _():
    db.create_all()


class U2FManager(registration.U2FRegistrationManager, verification.U2FSigningManager):

    def __init__(self):
        self.app_id = 'https://localhost:5000'

    def create_device_registration_model(self, *, transports, **kwargs):
        device = Device()
        device.u2f_transports = transports
        for k, v in kwargs.items():
            setattr(device, k, v)
        db.session.add(device)
        db.session.commit()
        return device

    def update_device_registration_counter(self, device, counter):
        device.counter = counter
        return device


u2f_manager = U2FManager()


@app.route("/")
def hello():
    dat = '<table>'
    dat += '<tr><th>Username</th><th>Registered keys</th></tr>'
    for user in User.query.all():
        dat += '<tr><td>' + user.name + '</td><td><table>\n'
        dat += '<tr><th>Index</th><th>Key Handle</th><th>Counter</th></tr>\n'
        for idx, key in enumerate(user.devices):
            dat += '<tr><td>{0}</td><td>{1.key_handle}</td><td>{1.counter}</td></tr>\n'.format(idx, key)
        dat += '</table></td></tr>'
    dat += '</table>'

    pre = ''
    if request.args.get('login', None) == 'success':
        pre = '<h1>You logged in successfully!!!!!!!!!!!!!!!</h1>'

    return pre + """
    <form method='POST' action='/register' >
    <input type='text' name='name' value='admin' />
    <input type='submit' value='Register new Device' />
    </form>
    <form method='POST' action='/login' >
    <input type='text' name='name' value='admin' />
    <input type='submit' value='Verify registered device' />
    </form>
    """ + dat


@app.route("/register.js", methods=['GET'])
def get_register_js():
    return (pathlib.Path(__file__).parent / 'register.js').open('r').read()


@app.route("/register", methods=['POST'])
def do_register_start():
    user_name = request.form['name']
    session['user_name'] = user_name
    user = User.query.filter_by(name=user_name).first()
    if not user:
        user = User(name=user_name)
        db.session.add(user)
        db.session.commit()
    data = u2f_manager.create_registration_challenge(session, user.devices)

    js = """
    const appId = {appId};
    const registerRequests = {registerRequests};
    const registeredKeys = {registeredKeys};
    """.format(
        appId=json.dumps(data['appId']),
        registerRequests=json.dumps(data['registerRequests']),
        registeredKeys=json.dumps(data['registeredKeys'])
    )
    return (
        '<pre>' + js + '</pre><script>' + js + '</script>' +
        '<span id="u2f_status">Loading</span><br /><a href="/">Back</a>' +
        '<form id="u2f_data" method="POST" action="/register2">'+
        '<input name="data" id="data" />'+
        '</form>'
        '<script src="register.js"></script>'
    )

@app.route("/register2", methods=['POST'])
def do_register_verify():
    user_name = session['user_name']
    user = User.query.filter_by(name=user_name).first()
    if not user:
        return 'No user by that name; <a href="/">Back</a>'
    data = request.json or json.loads(request.form['data'])
    device = u2f_manager.process_registration_response(
        session, data
    )
    device.user = user
    db.session.commit()
    return redirect('/')


@app.route("/login.js", methods=['GET'])
def get_login_js():
    return (pathlib.Path(__file__).parent / 'login.js').open('r').read()

@app.route("/login", methods=['POST'])
def do_login_start():
    user_name = request.form['name']
    session.user_name = user_name
    user = User.query.filter_by(name=user_name).first()
    if not user:
        return 'No user by that name; register a device first? <a href="/">Back</a>'
    if not user.devices:
        return 'No devices for that user; register a device first? <a href="/">Back</a>'

    devices = user.devices
    data = u2f_manager.create_signing_challenge(session, devices)
    js = """
    const appId = {appId};
    const challenge = {challenge};
    const registeredKeys = {registeredKeys};
    """.format(
        appId=json.dumps(data['appId']),
        challenge=json.dumps(data['challenge']),
        registeredKeys=json.dumps(data['registeredKeys'])
    )
    return (
        '<pre>' + js + '</pre><script>' + js + '</script>' +
        '<span id="u2f_status">Loading</span><br /><a href="/">Back</a>' +
        '<form id="u2f_data" method="POST" action="/login2">'+
        '<input name="data" id="data" />'+
        '</form>'
        '<script src="login.js"></script>'
    )


@app.route("/login2", methods=['POST'])
def do_login_verify():
    user_name = session['user_name']
    user = User.query.filter_by(name=user_name).first()
    if not user:
        return 'No user by that name; <a href="/">Back</a>'
    data = request.json or json.loads(request.form['data'])
    device = u2f_manager.process_signing_response(
        session, data, user.devices
    )
    db.session.commit()
    return redirect('/?login=success')



from .. import utils


def test_sha_256():
    assert utils.sha_256(
        b'') == (b'\xe3\xb0\xc4\x42\x98\xfc\x1c\x14'
                 b'\x9a\xfb\xf4\xc8\x99\x6f\xb9\x24'
                 b'\x27\xae\x41\xe4\x64\x9b\x93\x4c'
                 b'\xa4\x95\x99\x1b\x78\x52\xb8\x55')
    assert utils.sha_256(
        b'fido') == (b'\x04\xb1\xff\x4c\x19\x33\x58\xf9'
                     b'\x24\xef\xfd\xb5\x4e\xb6\xd2\x37'
                     b'\xfb\x49\x55\xe9\xd1\x43\xd9\x82'
                     b'\xf1\xf8\x63\x20\x3f\x18\x3f\x63')


def test_pop_bytes():
    arr = bytearray(b'0123456789')
    assert utils.pop_bytes(arr, 1) == b'0'
    assert arr == b'123456789'


class Test_parse_tlv_encoded_length():

    def test_zero_length(self):
        arr = bytearray(b'f\x00')
        assert utils.parse_tlv_encoded_length(arr) == 2

    def test_zero_length_with_flag(self):
        arr = bytearray(b'f\x80')
        assert utils.parse_tlv_encoded_length(arr) == 2

    def test_length_short_form(self):
        arr = bytearray(b'f\x05')
        assert utils.parse_tlv_encoded_length(arr) == (2 + 5)

    def test_length_long_form(self):
        # 0x80 => Long form
        # 0x01 => length of bytes representing message length
        arr = bytearray(b'f\x81\x05')
        assert utils.parse_tlv_encoded_length(arr) == (2 + 1 + 5)
    def test_length_really_long_form(self):
        # 0x80 => Long form
        # 0x7f => length of bytes representing message length
        arr = bytearray(b'f\xff' + (b'\0' * 0x7e) + b'\x05')
        assert utils.parse_tlv_encoded_length(arr) == (2 + 0x7f + 5)



from enum import Enum, unique

from . import _typing as typ


@unique
class U2FTransport(Enum):
    BLUETOOTH_RADIO = (0x80, 'br')
    BLUETOOTH_LOW_ENERGY_RADIO = (0x40, 'ble')
    USB = (0x20, 'usb')
    NFC = (0x10, 'nfc')
    USB_INTERNAL = (0x08, 'usb-internal')

    # def __init__(self, value: int, internal_name: str) -> None:
    #     self.value = value
    #     self.internal_name = internal_name

    @property
    def value(self):
        return super().value[0]

    @property
    def internal_name(self):
        return super().value[1]

    @staticmethod
    def from_byte(byte: int) -> typ.Collection['U2FTransport']:
        return [t for t in U2FTransport if t.value & byte]

    @staticmethod
    def to_byte(transports: typ.Collection['U2FTransport']) -> int:
        value = 0
        for transport in transports:
            value |= transport.value
        return value

    @staticmethod
    def _to_internal_int(
        transports: typ.Optional[typ.Collection['U2FTransport']],
    ) -> int:
        if transports is None:
            return -1
        else:
            return U2FTransport.to_byte(transports)

    @staticmethod
    def _from_internal_int(
        transports: int,
    ) ->typ.Optional[typ.Collection['U2FTransport']]:
        if transports < 0:
            return None
        else:
            return U2FTransport.from_byte(transports)


U2FTransports = typ.Optional[typ.Collection[U2FTransport]]


@unique
class RequestType(Enum):
    REGISTER = 'navigator.id.finishEnrollment'
    SIGN = 'navigator.id.getAssertion'



import abc
import struct

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import load_der_public_key

from . import _typing as typ
from .constants import PUB_KEY_DER_PREFIX
from .device import DeviceRegistration, device_as_client_dict
from .enums import RequestType
from .exceptions import U2FInvalidDataException, U2FStateException
from .utils import get_random_challenge, pop_bytes, sha_256
from .utils import validate_client_data, websafe_decode, websafe_encode


class U2FSigningManager(abc.ABC):
    """
    An abstract class that handles verifying a user's U2F token.

    Implementers must override ``update_device_registration_counter`` to
    perform any neccessary operations for storing the last seen device counter
    in a persistant storage.

    This class has 2 externally useable API methods
    ``create_signing_challenge`` and ``process_signing_response`` which
    should be used to provide the U2F verification/signing flow for a user.
    """

    SIGNING_SESSION_KEY = 'u2f_signing_challenge'

    def __init__(self, app_id: str) -> None:
        """
        Create a signing manager.


        """
        self.app_id = app_id

    @abc.abstractmethod
    def update_device_registration_counter(
            self,
            *,
            device: DeviceRegistration,
            counter: int
    ) -> DeviceRegistration:
        ...

    def filter_devices_by_app_id(
            self,
            registered_devices: typ.Collection[DeviceRegistration],
    ) -> typ.Collection[DeviceRegistration]:
        return [
            device for device in registered_devices
            if device.app_id == self.app_id
        ]

    def create_signing_challenge(
            self,
            session: typ.MutableMapping[str, typ.Any],
            registered_devices: typ.Collection[DeviceRegistration],
    ) -> typ.Mapping[str, typ.Any]:
        registered_devices = self.filter_devices_by_app_id(registered_devices)
        if not registered_devices:
            raise ValueError('Cannot issue a signing request with no keys.')
        challenge = websafe_encode(get_random_challenge())
        session[self.SIGNING_SESSION_KEY] = challenge
        keys = [device_as_client_dict(key) for key in registered_devices]
        return {
            'appId': self.app_id,
            'challenge': challenge,
            'registeredKeys': keys,

        }

    def process_signing_response(
            self,
            session: typ.MutableMapping[str, typ.Any],
            response_dict: typ.Mapping[str, str],
            registered_devices: typ.Collection[DeviceRegistration] = (),
    ) -> DeviceRegistration:
        registered_devices = self.filter_devices_by_app_id(registered_devices)
        key_handle = websafe_decode(response_dict.get('keyHandle', ''))
        challenge = session.get(self.SIGNING_SESSION_KEY, '')
        if not challenge:
            raise U2FStateException('Session missing required key.')
        device = self.get_key_by_handle(registered_devices, key_handle)
        signature_data = self.verify_signature_data(response_dict, challenge,
                                                    device)
        # Only update the counter once we've verified the device.
        counter = signature_data.counter
        return self.update_device_registration_counter(
            device=device,
            counter=counter
        )

    def get_key_by_handle(
            self,
            registered_keys: typ.Collection[DeviceRegistration],
            key_handle: str,
    ) -> DeviceRegistration:
        for key in registered_keys:
            if key.key_handle == key_handle:
                return key
        raise U2FInvalidDataException('Given key not found')

    def verify_signature_data(
            self,
            response_dict: typ.Mapping[str, str],
            challenge: str,
            device: DeviceRegistration,
    ) -> 'SignatureData':
        try:
            signature_data = SignatureData.from_base64(
                response_dict.get('signatureData', ''))
        except (ValueError, IndexError) as e:
            raise U2FInvalidDataException('Invalid signing data.') from e
        # Client data comes in as base64(usually?), so we standardise it
        #  into a decoded *string*. We then take the hash of that string
        #  for the verification step.
        client_data = validate_client_data(
            response_dict.get('clientData', ''),
            RequestType.SIGN,
            self.app_id,
            challenge,
        )
        challenge_param = sha_256(client_data.encode('utf-8'))
        app_param = sha_256(self.app_id.encode('idna'))
        signature_data.verify(
            app_param,
            challenge_param,
            device.public_key,
        )
        return signature_data


class SignatureData():
    @classmethod
    def from_base64(
            cls,
            base64_data: typ.Union[str, bytes],
    ) -> 'SignatureData':
        return cls(websafe_decode(base64_data))  # type: ignore

    def __init__(self, data: bytes) -> None:
        # https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.pdf
        buf = bytearray(data)
        self.user_presence = buf.pop(0)
        self.counter = struct.unpack('>I', pop_bytes(buf, 4))[0]
        self.signature = bytes(buf)

    def verify(self, app_param: bytes, chal_param: bytes, der_pubkey: bytes):
        pubkey = load_der_public_key(PUB_KEY_DER_PREFIX + der_pubkey,
                                     default_backend())
        verifier = pubkey.verifier(self.signature, ec.ECDSA(hashes.SHA256()))
        verifier.update(
            app_param + bytes([self.user_presence]) + struct.pack(
                '>I', self.counter) + chal_param, )
        try:
            verifier.verify()
        except InvalidSignature as e:
            raise U2FInvalidDataException(
                'Attestation signature is invalid') from e



#!/usr/bin/env python3
"""The setup script."""

from setuptools import find_packages, setup

requirements = [
    'cryptography>=2.3,<3',
]

flask_sample_requires = [
    'flask',
    'Flask-SQLAlchemy',
]


packages = find_packages(
    where='./',
    include=['fido_u2f', 'fido_u2f.*'],
)
if not packages:
    raise ValueError('No packages detected.')


with open('./README.rst', 'r') as readme_file:
    readme = readme_file.read()


setup(
    name='py-fido',
    version='0.3.0',
    description=(
        'A framework-agnostic implementation of the FIDO U2F server workflow'
    ),
    long_description=readme,
    author='Opal Symes',
    author_email='code@opal.codes',
    url='https://github.com/leesdolphin/py-fido/',
    packages=packages,
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'sample': flask_sample_requires,
    },
    zip_safe=False,
    package_data={
        'fido_u2f': ['py.typed'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Security',
    ],
)
